/*
  Implementation of arrays that can grow in size.

  Diego Zamboni, Mar 7, 1997.

  */

/* Chunk sizes in which the array grows */
#define EXT_ARRAY_CHUNK_SIZE  50

typedef struct ext_array_struct {
    int size;   /* Total number of units which have been allocated for
		   storage of the extendable array */
    int elems;  /* last_elem+1 */
    int last_elem; /* Largest index of all elements which have been set */
    void *data;
} extendable_array;

/* This is all implemented with macros to allow for specification of
   types. */

/* Initialize a variable of type extendable_array */
#define EXT_ARRAY_INIT(var,type,inisize)  \
{\
     var.size=inisize;\
     var.elems=0;\
     var.last_elem=-1;\
     var.data=(type *)malloc(sizeof(type) * inisize);\
}

/* Set the n-th element of the array to a value, extending the array
   if necessary. */
#define EXT_ARRAY_SET(var,type,n,val) \
{\
     if(n>=var.size) {\
	var.data=(type *)realloc(var.data,\
				 sizeof(type)*\
				 (((n/EXT_ARRAY_CHUNK_SIZE)+1)*\
				  EXT_ARRAY_CHUNK_SIZE));\
	var.size=(((n/EXT_ARRAY_CHUNK_SIZE)+1)*\
		  EXT_ARRAY_CHUNK_SIZE);\
     }\
     if(n>=var.elems) {\
	var.elems=n+1;\
     }\
     if(n>=var.last_elem) {\
	var.last_elem=n;\
     }\
     ((type *)var.data)[n]=val;\
}\

/* Get the value of an element. */
#define EXT_ARRAY_GET(var,type,n) (((type *)var.data)[n])
