/* ifile - intelligent mail filter for EXMH/MH
   Copyright (C) 1997  Jason Rennie <jr6b+@andrew.cmu.edu>
   
   irefile.c code Copyright (C) 1997 by Jason Rennie <jr6b+@andrew.cmu.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program (see file 'COPYING'); if not, write to the Free
   Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA  02111-1307, USA.
   */

/*
  irefile.c

  C implementation of irefile.perl.

  Written by Jason Rennie, beginning Mar 30, 1997.

  This is a fairly straight translation from the perl code.  The only purpose
  at this point was to increase the efficiency of the program.	Some code was
  taken from the ifilter.c code.  The design of the code is very similar to
  that of the irefile.perl code.

  This program relies on the availability of the MH refile program.

  */

#include <sys/stat.h>	   // stat functions
#include <unistd.h>	   // stat functions
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdarg.h>       /* va_*() */
#include <assert.h>       /* assert() */
#include <unistd.h>       /* pipe() fork() close() dup2() execlp() */
#include <errno.h>	  /* required by Solaris 5.5.1/gcc-2.7.2 for errno variable */
#include <ctype.h>        /* tolower() */
#include <string.h>

#include "assoc_array_int.h"

#include "error.h"        /* ifile message and error reporting library */
#include "ifile.h"        /* standard ifile library */

FILE * INFO = NULL;

/* Main program */
void main(int argc, char **argv) {
  /* To the extent that it was possible, the original variable names
     were kept to avoid confusion. */
  clock_t DMZ_start, DMZ_end;
  int i, j, k;
  char *home_dir = NULL;
  char tmpstr[BUFSIZ*8], *tmpstrptr = NULL;
  char *source_folder = NULL;
  char *source_file = NULL;
  char *dest_folder = NULL;
  extendable_array orig_message;
  char *acc_file = NULL;
  int tot_words;
  extendable_array tv;
  char *data_file = NULL;
  FILE *DATA = NULL;
  int tvsize;
  assoc_array_int n, docs, index, age, nk, sum_nk;
  int tot_docs;
  int tmpint;
  extendable_array word;
  char *curword = NULL;
  int cat_num, freq;
  char tmpstr2[BUFSIZ];
  extendable_array lines;
  char nonalpha[256];
  char *wrd = NULL;
  char *cur_wrd = NULL;
  int word_count;
  char * tmp;
  char * msg = NULL;
  char * msg_path = NULL;
  char * source_folder_path = NULL;
  FILE * MSG = NULL;
  int skip_msg = 0;
  extendable_array f_tv;
  int val;
  char * cur_word = NULL;
  int weight;
  mh_profile_struct mh_profile;

  /* Get initial time */
  DMZ_start = clock();
  
  INFO = ifile_open_log(argc, argv);

  /* Get home directory */
  home_dir = getenv("HOME");
  if (home_dir == NULL)
    ifile_error("Fatal: HOME environment variable not defined!\n");
  ifile_verbosify(ifile_verbose, "home directory = %s\n", home_dir);

  /* Reads in info from ~/.mh_profile */
  mh_profile = ifile_read_profile(home_dir);


  EXT_ARRAY_INIT(orig_message, char *, 10);
  
  ifile_verbosify(ifile_progress, "Reading command line arguments...\n");

  /* gets command line arguments */
  for (i=1; i < argc; i++)
    {
      if (strcmp(argv[i], "-src") == 0)
	{
	  if (argv[i+1][0] == '+')
	    {
	      /* skip '+' when creating SOURCE_FOLDER */
	      source_folder = ifile_sprintf("%s", argv[i+1] + 1);
	      i++;
	      ifile_verbosify(ifile_verbose, "source folder = %s\n", source_folder);
	    }
	  else
	    {
	      ifile_verbosify(ifile_silent, "Error: folder name %s lacks preceeding \'+\' or is in wrong format.\n", source_folder);
	      irefile_print_usage();
	    }
	}
      else if (strcmp(argv[i], "-file") == 0)
	{
	  source_file = ifile_sprintf("%s", argv[i+1]);
	  ifile_verbosify(ifile_verbose, "source file = %s\n", source_file);
	  i++;
	}
      else if ((strcmp(argv[i], "--help") == 0) ||
	       (strcmp(argv[i], "-h") == 0))
	irefile_print_usage();
      else if ((strcmp(argv[i], "-draft") == 0) ||	    
	       (strcmp(argv[i], "-nolink") == 0) ||
	       (strcmp(argv[i], "-preserve") == 0) ||
	       (strcmp(argv[i], "-nopreserve") == 0) ||
	       (strcmp(argv[i], "-rmmproc") == 0))
	ifile_verbosify(ifile_debug, "%s recognized as MH command line argument\n", argv[i]);
      else if (strcmp(argv[i], "-link") == 0)
	{
	  ifile_close_log();
	  ifile_mh_refile(argc, argv);   /* Normally does not exit */
	  ifile_error("Not able to call MH refile.\nDon't know what to do with -link option!\nNo refile action taken.\n");
	}
      else if (argv[i][0] == '+')
	{
	  dest_folder = ifile_sprintf("%s", argv[i]+1);
	  ifile_verbosify(ifile_verbose, "destination folder = %s\n", dest_folder);
	}
      else
	{
	  tmp = ifile_sprintf("%s", argv[i]);
	  j = orig_message.last_elem;
	  EXT_ARRAY_SET(orig_message, char *, j+1, tmp);
	  ifile_verbosify(ifile_verbose, "orig_message[%d] = %s\n", j+1, tmp);
	}
    }
  
  if (dest_folder == NULL)
    {
      ifile_verbosify(ifile_silent, "Error: No destination folder specified!\n");
      irefile_print_usage();
    }
  
  if (source_folder == NULL)
    {
      FILE * CONTEXT;

      if ((CONTEXT = fopen(mh_profile.context_file, "r")) == NULL)
	{
	  ifile_verbosify(ifile_silent, "No source folder given.  No context file found: %s!\nYou must specify a source folder!\n", mh_profile.context_file);
	  irefile_print_usage();
	}
      
      readline(tmpstr, sizeof(tmpstr), CONTEXT);
      while (!feof(CONTEXT))
	{
	  tmpstrptr = strtok(tmpstr, " \t");
	  if (strcmp(tmpstrptr, "Current-Folder:") == 0)
	    {
	      tmpstrptr = strtok(NULL, " \t");
	      source_folder = ifile_sprintf("%s", tmpstrptr);
	      ifile_verbosify(ifile_verbose, "current MH folder = %s\n", tmpstrptr);
	    }
	  readline(tmpstr, sizeof(tmpstr), CONTEXT);
	}
    }
  
  if (orig_message.last_elem == -1)
    {
      FILE * MH_SEQ;
      char * mh_seq = ifile_sprintf("%s/%s%s/.mh_sequences", home_dir,
				    mh_profile.mail_path, source_folder);
      if ((MH_SEQ = fopen(mh_seq, "r")) == NULL)
	{
	  ifile_verbosify(ifile_silent, "Could not open MH sequences file: %s.\nNot able to determine current message.\nPlease specify a message to be refiled\n", mh_seq);
	  irefile_print_usage();
	}
      
      readline(tmpstr, sizeof(tmpstr), MH_SEQ);
      while (!feof(MH_SEQ))
	{
	  tmpstrptr = strtok(tmpstr, " \t");
	  if (strcmp(tmpstrptr, "cur:") == 0)
	    {
	      tmpstrptr = strtok(NULL, " \t");
	      tmp = strdup(tmpstrptr);
	      EXT_ARRAY_SET(orig_message, char *, 0, tmp);
	      ifile_verbosify(ifile_verbose, "Current message is %s\n", tmp);
	      free(tmp);
	      tmp = NULL;
	    }
	  readline(tmpstr, sizeof(tmpstr), MH_SEQ);
	}
      free(mh_seq);
    }
  
  if (orig_message.last_elem == -1)
    {
      ifile_verbosify(ifile_verbose, "No cur message, no source message(s) specified\n");
      irefile_print_usage();
    }

  if (source_file != NULL)
    {
      if (orig_message.last_elem >= 0)
	{
	  ifile_verbosify(ifile_quiet, "Warning: Messages other than %s are being ignored.\n", source_file);
	  
	  /* no use for orig_message strings - free used memory */
	  for (i=0; i <= orig_message.last_elem; i++)
	    {
	      free(EXT_ARRAY_GET(orig_message, char *, i));
	      EXT_ARRAY_SET(orig_message, char *, i, NULL);
	    }
	  orig_message.elems = 0;
	  orig_message.last_elem = -1;
	}
      
      if (source_folder != NULL)
	{
	  ifile_verbosify(ifile_quiet, "Warning: \'-src\' command line option is being ignored\n");
	  free(source_folder);
	  source_folder = NULL;
	}
      /* set 0th orig_message equal to source_file */
      tmp = ifile_sprintf("%s", source_file);
      EXT_ARRAY_SET(orig_message, char *, 0, tmp);
      ifile_free(tmp);
    }

  acc_file = ifile_sprintf("%s/%s", home_dir, IFILE_ACCURACY);
  ifile_update_accuracy(acc_file, 0, orig_message.elems);
  
  /* Now read the database */
  tot_words = 0;
  
  EXT_ARRAY_INIT(tv, char *, EXT_ARRAY_CHUNK_SIZE);
  EXT_ARRAY_SET(tv,char *, 0, NULL);
  tvsize = 0;
  
  data_file = ifile_sprintf("%s/.idata", home_dir);

  DMZ_start=clock();
  ifile_verbosify(ifile_progress, "Opening %s for reading...\n", data_file);

  /* initialize variables to store info from database */
  assoc_array_int_init(n);
  assoc_array_int_init(docs);
  assoc_array_int_init(index);
  assoc_array_int_init(age);
  assoc_array_int_init(nk);
  assoc_array_int_init(sum_nk);
  EXT_ARRAY_INIT(word,char*,1000);
  
  if((DATA = fopen(data_file, "r")) != NULL)
    {
      /* Read the category names */
      readline(tmpstr, sizeof(tmpstr), DATA);
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  tvsize++;
	  EXT_ARRAY_SET(tv,char*,tvsize,strdup(tmpstrptr));
	  tmpstrptr = strtok(NULL, " \t");
	}
      
      /* Read total number of words for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      i = 1;
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  assoc_array_int_set(n, EXT_ARRAY_GET(tv,char*,i), atoi(tmpstrptr));
	  i++;
	  tmpstrptr=strtok(NULL, " \t");
	}
      if(i <= tvsize)
	{
	  ifile_error("Bad data file format - line #2\n");
	}
      
      /* Read number of documents for each category */
      readline(tmpstr, sizeof(tmpstr), DATA);
      tot_docs = 0;
      i = 1;
      tmpstrptr = strtok(tmpstr, " \t");
      while (tmpstrptr != NULL)
	{
	  tmpint = atoi(tmpstrptr);
	  assoc_array_int_set(docs, EXT_ARRAY_GET(tv, char*, i),tmpint);
	  i++;
	  tot_docs += tmpint;
	  tmpstrptr = strtok(NULL, " \t");
	}
      if (i <= tvsize)
	{
	  ifile_error("Bad data file format - line #3\n");
	}
      
      /* Read word frequencies */
      EXT_ARRAY_SET(word,char*,0,NULL);
      
      readline(tmpstr, sizeof(tmpstr), DATA);
      while (!feof(DATA))
	{
	  tmpstrptr = strtok(tmpstr, " \t");
	  if (tmpstrptr == NULL)
	    break;
	  tot_words++;
	  EXT_ARRAY_SET(word, char *, tot_words, strdup(tmpstrptr));
	  assoc_array_int_set(index, tmpstrptr, tot_words);
	  curword = tmpstrptr;
	  
	  /* Read age of word */
	  tmpstrptr = strtok(NULL, " \t");
	  assoc_array_int_set(age, curword, atoi(tmpstrptr));
	  
	  /* Read frequency per category cat_num:freq */
	  tmpstrptr = strtok(NULL, ":");
	  while (tmpstrptr != NULL)
	    {
	      cat_num = atoi(tmpstrptr);
	      tmpstrptr = strtok(NULL, " \t");
	      freq = atoi(tmpstrptr);
	      sprintf(tmpstr2, "%s,%s",
		      EXT_ARRAY_GET(word, char *, tot_words),
		      EXT_ARRAY_GET(tv, char *, cat_num));
	      assoc_array_int_set(nk, tmpstr2, freq);

	      /* sums frequencies per word for later usage */
	      assoc_array_int_set(sum_nk, EXT_ARRAY_GET(word, char *, tot_words), assoc_array_int_get(sum_nk, EXT_ARRAY_GET(word, char *, tot_words)) + freq);
	      tmpstrptr = strtok(NULL, ":");
	    }
	  readline(tmpstr, sizeof(tmpstr), DATA);
	}
    }
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Read %d categories, %d words.  Elapsed time: %.3f sec\n", tvsize, tot_words,
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  fclose(DATA);

  DMZ_start = clock();

  source_folder_path = ifile_sprintf("%s/%s%s/", home_dir,
				     mh_profile.mail_path, source_folder);
  ifile_verbosify(ifile_verbose, "source folder path = %s\n", source_folder_path);
  
  DMZ_start = clock();
  ifile_verbosify(ifile_progress, "Reading messages...\n");

  /* refiles messages one at a time */
  
  for (i=0; i <= orig_message.last_elem; i++)
    {
      skip_msg = FALSE;

      if (source_file == NULL)
	{
	  msg = ifile_sprintf("%s", EXT_ARRAY_GET(orig_message, char *, i));
	  msg_path = ifile_sprintf("%s%s", source_folder_path, msg);
	}
      else
	msg_path = strdup(source_file);

      ifile_verbosify(ifile_verbose, "Reading message: %s\n", msg_path);

      if (source_folder != NULL)
	{
	  if ((MSG = fopen(msg_path, "r")) == NULL)
	    {
	      ifile_verbosify(ifile_quiet, "Warning: Could not open %s\n",
			      msg_path);
	      skip_msg = TRUE;
	    }
	}
      else /* the entire message path was specified on the command line */
	{
	  if ((MSG = fopen(msg, "r")) == NULL)
	    {
	      ifile_verbosify(ifile_quiet, "Warning: Could not open %s\n", 
			      msg_path);
	      skip_msg = TRUE;
	    }
	}
      
      ifile_free(msg);
      
      /* Initialize the non-alphabetic-character token separator */
      k = 0;
      for (j = 1; j <= 255; j++) 
	{
	  if(!((j >= 'a' && j <= 'z') || (j >= 'A' && j <= 'Z')))
	    nonalpha[k++] = j;
	}
      nonalpha[k] = '\0';
      
      EXT_ARRAY_INIT(lines ,char *, EXT_ARRAY_CHUNK_SIZE);
      j = 0;
      word_count = 0;
      
      /* Read the message in an array for processing */
      readline(tmpstr, sizeof(tmpstr), MSG);
      while (!feof(MSG)) 
	{
	  weight = 1;
	  k = lines.last_elem + 1;
	  EXT_ARRAY_SET(lines, char*, k, strdup(tmpstr));
	  j++;
	  /* Split line in words and store them in another array */
	  tmpstrptr = strtok(tmpstr, nonalpha);
	  while (tmpstrptr != NULL)
	    {
	      wrd = tmpstrptr;
	      /* Skip tiny words */
	      if (strlen(wrd) > 2) 
		{
		  cur_wrd = wrd;
		  /* Convert to lower case */
		  while((*wrd) != '\0') 
		    {
		      (*wrd) = tolower(*wrd);
		      wrd++;
		    }
		  
		  /* Update vocab list and word statistics */
		  if (assoc_array_int_get(index, cur_wrd) == 0)
		    {
		      ifile_verbosify(ifile_debug, "Creating new entry for %s\n", cur_wrd);
		      tmp = strdup(cur_wrd);
		      k = word.last_elem + 1;
		      EXT_ARRAY_SET(word, char *, k, tmp);
		      assoc_array_int_set(index, cur_wrd, k);

		      tmp = ifile_sprintf("%s,%s", cur_wrd, dest_folder);
		      assoc_array_int_set(nk, tmp, weight + 
					  assoc_array_int_get(nk, tmp));
		      assoc_array_int_set(sum_nk, cur_wrd, assoc_array_int_get(sum_nk, cur_wrd) + weight);
		    }
		  else
		    {	
		      tmp = ifile_sprintf("%s,%s", cur_wrd, dest_folder);
		      assoc_array_int_set(nk, tmp, weight + 
					  assoc_array_int_get(nk, tmp));
		      assoc_array_int_set(sum_nk, cur_wrd, assoc_array_int_get(sum_nk, cur_wrd) + weight);
		      ifile_free(tmp);
		    }
		  assoc_array_int_set(n, dest_folder, weight +
				      assoc_array_int_get(n, dest_folder));
		  
		  tmp = ifile_sprintf("%s,%s", cur_wrd, dest_folder);
		  ifile_free(tmp);
		}
	      tmpstrptr = strtok(NULL, nonalpha);
	    }
	  readline(tmpstr, sizeof(tmpstr), MSG);
	}
      
      fclose(MSG);
      
      assoc_array_int_set(docs, dest_folder, 1 +
			  assoc_array_int_get(docs, dest_folder));	
      for (j=1; j <= word.last_elem; j++)
	assoc_array_int_set(age, EXT_ARRAY_GET(word, char *, j), 1 +
			    assoc_array_int_get(age, EXT_ARRAY_GET(word, char *, j)));
      
      
      DMZ_end = clock();
      ifile_verbosify(ifile_verbose, "Finished reading message after %.3f sec.\n",
		      ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
    }
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Finishing reading messages.  Elapsed time: %.3f sec\n",
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);

  /* Reopen and write the learned information */
  
  
  data_file = ifile_sprintf("%s/.idata", home_dir);
  
  DMZ_start = clock();
  ifile_verbosify(ifile_progress, "Writing %s to disk...\n", data_file);

  EXT_ARRAY_INIT(f_tv, char *, 50);
  
  if ((DATA = fopen(data_file, "w")) != NULL) {
    
    /* Reads current folder list from "~/[mail_path]/.folders", keeps the
       names in memory and writes them to disk as the first line of .idata */
    
    char * folders_file = NULL;
    FILE * FOLDERS = NULL;
    
    folders_file = ifile_sprintf("%s/%s.folders", home_dir,
				 mh_profile.mail_path);
    
    EXT_ARRAY_INIT(f_tv, char *, 50);
    
    if ((FOLDERS = fopen(folders_file, "r")) != NULL)
      {
	for (i = 1; fgets(tmpstr, sizeof(tmpstr), FOLDERS); i++)
	  {
	    if (tmpstr[strlen(tmpstr)-1] == '\n')
	      tmpstr[strlen(tmpstr)-1] = '\0';
	    EXT_ARRAY_SET(f_tv, char *, i, malloc(strlen(tmpstr)+1));
	    strcpy(EXT_ARRAY_GET(f_tv, char *, i), tmpstr);
	    fputs(tmpstr, DATA);
	    fputc(' ', DATA);
	  }
      }
    else
      /* Umm... this is a really BAD way to handle not being able to
	 find the .folders file.  I need to fix this in the future */
      ifile_error("Could not locate .folders file\n");
    
    fputc('\n', DATA);
    
    /* Writes to .idata the total number of words observed for each of the
       folders */
    
    for (i=1; i <= f_tv.last_elem; i++)
      {
	val = assoc_array_int_get(n, EXT_ARRAY_GET(f_tv, char *, i));
	if (val > 0)
	  fprintf(DATA, "%d ", val);
	else
	  fprintf(DATA, "0 ");
      }
    
    fputc('\n', DATA);
    
    /* Writes to .idata the total number of documents observed for each of
       the folders */
    
    for (i=1; i <= f_tv.last_elem; i++)
      {
	val = assoc_array_int_get(docs, EXT_ARRAY_GET(f_tv, char *, i));
	if (val > 0)
	  fprintf(DATA, "%d ", val);
	else
	  fprintf(DATA, "0 ");
      }
    
    fputc('\n', DATA);
    
    /* For each word in the vocabulary, writes frequency by folder if the
       word is frequent enough to warrant recording */
    
    for (i=1; i <= word.last_elem; i++)
      {
	cur_word = EXT_ARRAY_GET(word, char *, i);
	/* if the word is infrequent, discard it (don't print it to DATA) */
	if ((assoc_array_int_get(sum_nk, cur_word) > 3) ||
	    (assoc_array_int_get(age, cur_word) < 16))
	  {
	    fprintf(DATA, "%s %d ", cur_word,
		    assoc_array_int_get(age, cur_word));

	    for (j=1; j <= f_tv.last_elem; j++)
	      {
		sprintf(tmpstr, "%s,%s",
			EXT_ARRAY_GET(word, char *, i),
			EXT_ARRAY_GET(f_tv, char *, j));
		val = assoc_array_int_get(nk, tmpstr);
		if (val > 0)
		  fprintf(DATA, "%d:%d ", j, val);
	      }
	    fputc('\n', DATA);
	  }
      }
  }
  fclose(DATA);
  
  DMZ_end = clock();
  ifile_verbosify(ifile_progress, "Wrote %d categories, %d words.  Elapsed time: %.3f sec\n", tvsize, tot_words,
		  ((float)(DMZ_end-DMZ_start))/CLOCKS_PER_SECOND);
  
  /* if refile.bak is executable, then use it to refile all the messages */
  if (ifile_exe(MH_REFILE))
    ifile_mh_refile(argc, argv);
  
  ifile_error("Error: Not able to properly refile messages with MH refile!\n");
}
